part of nueli.core.config;

/// Photo shortcut
Photo get $photo => Photo();

/// Photo ~ [Image] and [Icon] configuration class.
class Photo extends Config<Photo, PhotoData> {
  Color? _color;
  Alignment? _align;
  Ink? _ink;

  Photo() : super(PhotoData(blendMode: BlendMode.srcATop));

  Photo _config({
    Color? color,
    Alignment? align,
    Ink? ink,
  }) {
    _color = color ?? _color;
    _align = align ?? _align;
    _ink = ink ?? _ink;
    return this;
  }

  // variant commands
  Photo color(Color color) => _config(color: color)..data.blendMode = BlendMode.srcIn;
  Photo align(Alignment align) => _config(align: align);

  // fixed commands
  Photo withSize(double width, [double? height]) => this & Size(width, height ?? width);
  Photo withPicture(Picture? photo) => this & photo;
  Photo withBlendMode(BlendMode? blendMode) => this & blendMode;
  Photo withBoxFit(BoxFit? boxFit) => this & boxFit;

  // config commands
  Photo $ink(Ink ink) => _config(ink: ink);

  // shortcuts
  Photo get color_none => color(palette.none); // ignore: non_constant_identifier_names
  Photo get color_primary => color(palette.primary); // ignore: non_constant_identifier_names
  Photo get color_accent => color(palette.accent); // ignore: non_constant_identifier_names
  Photo get color_background => color(palette.background); // ignore: non_constant_identifier_names
  Photo get color_onBackground => color(palette.onBackground); // ignore: non_constant_identifier_names

  Photo get align_start => align(Alignment.topLeft); // ignore: non_constant_identifier_names
  Photo get align_center => align(Alignment.center); // ignore: non_constant_identifier_names
  Photo get align_end => align(Alignment.bottomRight); // ignore: non_constant_identifier_names

  @override
  Photo operator *(double factor) {
    final photoAp = _PhotoApprox(this);
    return _config(
      color: photoAp.color * factor,
      align: photoAp.align * factor,
    );
  }

  @override
  Photo operator -(Photo operand) {
    final photoAp = _PhotoApprox(this);
    final photoAp2 = _PhotoApprox(operand);
    return _config(
      align: photoAp.align - photoAp.align,
      color: photoAp.color - photoAp2.color,
    );
  }

  @override
  Photo operator +(Photo operand) {
    final photoAp = _PhotoApprox(this);
    final photoAp2 = _PhotoApprox(operand);
    return _config(
      align: photoAp.align + photoAp.align,
      color: photoAp.color + photoAp2.color,
    );
  }

  @override
  Photo lerp(ico) {
    if (ico is Photo)
      _config(
        align: ico._align ?? _align,
        color: ico._color ?? _color,
      );
    return super.lerp(ico);
  }

  Picture get toPhotoData => data.photo ?? Picture.library(CupertinoIcons.exclamationmark_circle);

  @override
  Widget call() {
    final photoAp = _PhotoApprox(this);

    /// library
    if (toPhotoData.data != null) {
      final photo = Icon(
        toPhotoData.data,
        color: photoAp.color,
        size: data.shortestSide,
      );
      if (_ink == null) return photo;
      return (_ink! & Radii.all(data.longestSide) & photo | InkType.response).call();
    }

    /// asset
    if (toPhotoData.path != null) {
      final photo = Image.asset(
        toPhotoData.path!,
        fit: data.boxFit,
        alignment: photoAp.align,
        color: photoAp.color,
        colorBlendMode: data.blendMode,
        width: data.width,
        height: data.height,
      );
      if (_ink == null) return photo;
      return (_ink! & Radii.all(data.longestSide) & photo | InkType.response).call();
    }

    /// network
    if (toPhotoData.src != null) {
      Widget loader(ImageChunkEvent chunk) {
        final loadValue = (chunk.expectedTotalBytes != null ? chunk.cumulativeBytesLoaded / chunk.expectedTotalBytes! : null);
        return Center(
          child: Container(
            height: 36,
            width: 36,
            padding: const EdgeInsets.all(8.0),
            child: CircularProgressIndicator(
              strokeWidth: 2,
              backgroundColor: palette.background,
              value: loadValue,
            ),
          ),
        );
      }

      Widget error([StackTrace? trace]) {
        final errType = (data.width ?? 0) > 100 && (data.height ?? 0) > 50;
        final errPhoto = $photo.color(palette.error) & CupertinoIcons.exclamationmark_circle;
        final errText = $txt.color((_color ?? palette.onBackground).withAlpha(100)) & 'An error occurred' & TextAlign.center;
        return Center(child: Padding(padding: const EdgeInsets.all(8.0), child: errType ? errText() : errPhoto()));
      }

      final photo = Image.network(
        toPhotoData.src!,
        color: photoAp.color,
        fit: data.boxFit,
        alignment: photoAp.align,
        colorBlendMode: data.blendMode,
        height: data.height,
        width: data.width,
        errorBuilder: (c, o, s) => error(s),
        loadingBuilder: (c, child, chunk) => chunk == null ? child : loader(chunk),
      );
      if (_ink == null) return photo;
      return (_ink! & Radii.all(data.longestSide) & photo | InkType.response).call();
    }

    return SizedBox();
  }
}

class PhotoData extends ConfigData<PhotoData> {
  Picture? photo;
  double? width;
  double? height;
  BlendMode? blendMode;
  BoxFit? boxFit;

  PhotoData({
    Widget? child,
    this.photo,
    this.width,
    this.height,
    this.blendMode,
    this.boxFit,
  }) : super(child);

  double get shortestSide => min(width ?? 0, height ?? 0);
  double get longestSide => max(width ?? 0, height ?? 0);

  PhotoData lerp(data) {
    if (data is IconData) photo = Picture.library(data);
    if (data is Picture) photo = data;

    if (data is Size) {
      width = data.width;
      height = data.height;
    }

    if (data is BlendMode) blendMode = data;
    if (data is BoxFit) boxFit = data;

    if (data is PhotoData) {
      photo = data.photo;
      width = data.width;
      height = data.height;
      blendMode = data.blendMode;
      boxFit = data.boxFit;
    }
    return super.lerp(data);
  }
}

class PhotoTween extends Tween<Photo> {
  PhotoTween({Photo? begin, Photo? end}) : super(begin: begin, end: end);

  @override
  Photo transform(double value) => (begin ?? $photo) + ((end ?? $photo) - (begin ?? $photo)) * value;
}

class PhotoTransition extends AnimatedWidget {
  final Widget? child;
  PhotoTransition({required Animation<Photo> photo, this.child}) : super(listenable: photo);
  Animation<Photo> get photo => listenable as Animation<Photo>;

  @override
  Widget build(BuildContext context) {
    return AnimatedBuilder(
      animation: listenable,
      builder: (c, child) => photo.value.withChild(child)(),
    );
  }
}

class _PhotoApprox {
  final Photo data;
  _PhotoApprox(this.data);

  Color get color => data._color ?? palette.none;
  Alignment get align => data._align ?? Alignment.center;
}
