part of nueli.core.config;

/// X3D shortcut
X3D get $3D => X3D();

/// X3D ~ [Matrix4] configuration class.
class X3D extends Config<X3D, X3DData> {
  Alignment? _alignment;
  List<_TransformData> _transforms = [];

  X3D() : super(X3DData());

  X3D _config({
    Alignment? alignment,
    List<_TransformData> transforms = const [],
  }) {
    _alignment = alignment ?? _alignment;
    _transforms = _transforms..addAll(transforms);
    return this;
  }

  // variant commands
  X3D align(Alignment alignment) => _config(alignment: alignment);
  X3D skew({double x = 0, double y = 0, double z = 0}) => (skew_x(x)..skew_y(y))..rotate(z);
  X3D zoom(double zoom) => _config(transforms: [_TransformData(zoom, (t, d) => t..rotateZ(d))]);
  X3D translate({double x = 0, double y = 0, double z = 0}) => (translate_x(x)..translate_y(y))..translate_z(z);
  X3D shift({double x = 0, double y = 0, double z = 0}) => (shift_x(x)..shift_y(y))..shift_z(z);
  X3D scale({double x = 0, double y = 0, double z = 0}) => (scale_x(x)..scale_y(y))..scale_z(z);

  // fixed commands
  X3D withBase(Matrix4? base) => this & base;

  // shortcuts
  X3D skew_x(double x) => _config(transforms: [_TransformData(x, (t, d) => t..rotateX(d))]); // ignore: non_constant_identifier_names
  X3D skew_y(double y) => _config(transforms: [_TransformData(y, (t, d) => t..rotateY(d))]); // ignore: non_constant_identifier_names
  X3D rotate(double z) => _config(transforms: [_TransformData(z, (t, d) => t..rotateZ(d))]); // ignore: non_constant_identifier_names
  X3D translate_x(double x) => _config(transforms: [_TransformData(x, (t, d) => t..translate(d))]); // ignore: non_constant_identifier_names
  X3D translate_y(double y) => _config(transforms: [_TransformData(y, (t, d) => t..translate(0, d))]); // ignore: non_constant_identifier_names
  X3D translate_z(double z) => _config(transforms: [_TransformData(z, (t, d) => t..translate(0, 0, d))]); // ignore: non_constant_identifier_names
  X3D shift_x(double x) => _config(transforms: [_TransformData(x, (t, d) => t..leftTranslate(d))]); // ignore: non_constant_identifier_names
  X3D shift_y(double y) => _config(transforms: [_TransformData(y, (t, d) => t..leftTranslate(0, d))]); // ignore: non_constant_identifier_names
  X3D shift_z(double z) => _config(transforms: [_TransformData(z, (t, d) => t..leftTranslate(0, 0, d))]); // ignore: non_constant_identifier_names
  X3D scale_x(double x) => _config(transforms: [_TransformData(x, (t, d) => t..scale(d))]); // ignore: non_constant_identifier_names
  X3D scale_y(double y) => _config(transforms: [_TransformData(y, (t, d) => t..scale(0, d))]); // ignore: non_constant_identifier_names
  X3D scale_z(double z) => _config(transforms: [_TransformData(z, (t, d) => t..scale(0, 0, d))]); // ignore: non_constant_identifier_names

  @override
  X3D operator *(double factor) {
    final x3DAp = _X3DApprox(this);
    return _config(
      alignment: x3DAp.alignment * factor,
      transforms: _transforms.map((e) => e * factor).toList(),
    );
  }

  @override
  X3D operator -(X3D operand) {
    final x3DAp = _X3DApprox(this);
    final x3DAp2 = _X3DApprox(operand);
    List<_TransformData> transforms = _transforms.minus(operand._transforms);
    return X3D() & data
      .._config(alignment: x3DAp.alignment - x3DAp2.alignment, transforms: transforms);
  }

  @override
  X3D operator +(X3D operand) {
    final x3DAp = _X3DApprox(this);
    final x3DAp2 = _X3DApprox(operand);
    List<_TransformData> transforms = _transforms.plus(operand._transforms);
    return _config(alignment: x3DAp.alignment + x3DAp2.alignment, transforms: transforms);
  }

  @override
  X3D lerp(matrix) {
    if (matrix is X3D)
      _config(
        alignment: matrix._alignment ?? _alignment,
        transforms: matrix._transforms + _transforms,
      );
    return super.lerp(matrix);
  }

  Matrix4 get toMatrix4 {
    Matrix4 mat = data.base;
    for (_TransformData transform in _transforms) mat = transform(mat);
    return mat;
  }

  Widget call() {
    _alignment = _alignment ?? Alignment.center;
    return Transform(child: data.child, alignment: _alignment, transform: toMatrix4);
  }
}

class X3DData extends ConfigData<X3DData> {
  late Matrix4 base;

  X3DData({Widget? child, Matrix4? base}) : super(child) {
    base = Matrix4.identity()..setEntry(3, 2, 0.001);
  }

  X3DData lerp(data) {
    if (data is Matrix4) base = data;

    if (data is X3DData) base = data.base;
    return super.lerp(data);
  }
}

class X3DTween extends Tween<X3D> {
  X3DTween({X3D? begin, X3D? end}) : super(begin: begin, end: end);

  @override
  X3D transform(double value) => (begin ?? $3D) + ((end ?? $3D) - (begin ?? $3D)) * value;
}

class X3DTransition extends AnimatedWidget {
  final Widget? child;
  X3DTransition({required Animation<X3D> x3D, this.child}) : super(listenable: x3D);
  Animation<X3D> get x3D => listenable as Animation<X3D>;

  @override
  Widget build(BuildContext context) {
    return AnimatedBuilder(
      animation: listenable,
      builder: (c, child) => x3D.value.withChild(child)(),
    );
  }
}

class _X3DApprox {
  final X3D x3D;
  _X3DApprox(this.x3D);

  Alignment get alignment => x3D._alignment ?? Alignment.center;
  List<_TransformData> get transforms => x3D._transforms;
}

class _TransformData with Subtract<_TransformData>, Add<_TransformData> {
  double value;
  Matrix4 Function(Matrix4, double) transform;
  _TransformData(this.value, this.transform);
  Matrix4 call(Matrix4 matrix) => transform(matrix, value);

  _TransformData operator -(_TransformData element) {
    return _TransformData(value - element.value, (mat, d) => transform(element.transform(mat, d), d));
  }

  @override
  _TransformData operator +(_TransformData element) {
    return _TransformData(value + element.value, (mat, d) => transform(element.transform(mat, d), d));
  }

  _TransformData operator *(double factor) {
    return _TransformData(value + factor, (mat, d) => transform(mat, d));
  }
}
